package io.tswf.minecraft.blockventure.core.world.biome

import com.mojang.datafixers.util.Pair
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.transforms.validation.checkedmapconstructor.CheckedMapConstructor
import io.tswf.minecraft.blockventure.core.exception.DimensionBiomesEnrichException
import io.tswf.minecraft.blockventure.core.transmutation.view.ViewMultiNoiseBiomeSourceParameterList
import net.minecraft.registry.Registry
import net.minecraft.registry.RegistryKey
import net.minecraft.registry.entry.RegistryEntry
import net.minecraft.util.Identifier
import net.minecraft.world.biome.Biome
import net.minecraft.world.biome.source.MultiNoiseBiomeSourceParameterList
import net.minecraft.world.biome.source.util.MultiNoiseUtil

@CompileStatic
class DimensionBiomesEnrichService {

    @Lazy
    static List<IDimensionBiomesEnrichProvider> dimensionBiomesEnrichProviders = ServiceLoader.load(IDimensionBiomesEnrichProvider)
        .toList()

    static void enrichDimensionBiomesByRegistry(Registry<MultiNoiseBiomeSourceParameterList> registry) {
        registry.entrySet.stream()
            .forEach { dimensionNoiseBiomesEntry ->
                def dimensionKey = dimensionNoiseBiomesEntry.key.value

                for (def dimensionBiomesEnrichProvider : dimensionBiomesEnrichProviders) {
                    try {
                        def multiNoiseBiomeList = dimensionNoiseBiomesEntry.value
                        def viewMultiNoiseBiomeList = (ViewMultiNoiseBiomeSourceParameterList) multiNoiseBiomeList
                        def dimensionBiomesEnrichContext = new DimensionBiomesEnrichContext(
                                dimensionTypeId: dimensionKey
                        )

                        dimensionBiomesEnrichProvider.enrichDimensionBiomes(dimensionBiomesEnrichContext)

                        dimensionBiomesEnrichContext.biomesToRemove?.let { biomesToRemoveList ->
                            if (biomesToRemoveList.isNullOrEmpty()) return

                            def newEntries = multiNoiseBiomeList.entries.entries.stream()
                                    .filter { it.second !in biomesToRemoveList }
                                    .toList()
                                    .with {
                                        new MultiNoiseUtil.Entries(it)
                                    }

                            viewMultiNoiseBiomeList.entries = newEntries
                        }

                        dimensionBiomesEnrichContext.biomeKeysToRemove?.let { biomeKeysToRemoveList ->
                            if (biomeKeysToRemoveList.isNullOrEmpty()) return

                            def (regEntries, directEntries) = multiNoiseBiomeList.entries.entries.splitBy { it.second.key.isPresent() }

                            def newEntries = regEntries
                                    .filter { it.second.key.orNull() !in biomeKeysToRemoveList }
                                    .toList()
                                    .with {
                                        new MultiNoiseUtil.Entries(it + directEntries)
                                    }

                            viewMultiNoiseBiomeList.entries = newEntries
                        }

                        dimensionBiomesEnrichContext.sameBiomesToAdd?.let { biomesToAddMap ->
                            if (biomesToAddMap.isNullOrEmpty()) return

                            def oldBiomeEntries = multiNoiseBiomeList.entries.entries

                            biomesToAddMap.entrySet().forEach { biomesToAddEntry ->
                                def hypercubesToAdd = oldBiomeEntries.stream()
                                        .filter { it.second.key.orNull() == biomesToAddEntry.key }
                                        .map { it.first }
                                        .map {
                                            new MultiNoiseUtil.NoiseHypercube(
                                                    it.temperature(),
                                                    it.humidity(),
                                                    it.continentalness(),
                                                    it.erosion(),
                                                    it.depth(),
                                                    it.weirdness(),
                                                    it.offset()
                                            )
                                        }
                                        .toSet()

                                def pairsToAdd = hypercubesToAdd.stream()
                                        .flatMap { cube ->
                                            biomesToAddEntry.value.stream()
                                                    .map { biome -> Pair.of(cube, biome) }
                                        }
                                        .distinct()
                                        .toSet()

                                def newEntriesList = oldBiomeEntries + pairsToAdd
                                def newEntries = new MultiNoiseUtil.Entries(newEntriesList)
                                viewMultiNoiseBiomeList.entries = newEntries
                                return
                            }
                        }
                    } catch (Throwable e) {
                        throw new DimensionBiomesEnrichException(e, dimensionKey, dimensionBiomesEnrichProvider)
                    }
                }
            }
    }

}

@CompileStatic
@CheckedMapConstructor
class DimensionBiomesEnrichContext {
    Identifier dimensionTypeId

    Map<RegistryKey<Biome>, List<RegistryEntry<Biome>>> sameBiomesToAdd = new HashMap<RegistryKey<Biome>, List<RegistryEntry<Biome>>>()
        .withDefault {
            (List<RegistryEntry<Biome>>) new ArrayList()
        }

    List<RegistryEntry<Biome>> biomesToRemove = new ArrayList<>()
    List<RegistryKey<Biome>> biomeKeysToRemove = new ArrayList<>()
    
    void addSameBiome(RegistryKey<Biome> sameAs, Biome biome) {
        addSameBiome(sameAs, biome.toRegistryEntry())
    }

    void addSameBiome(RegistryKey<Biome> sameAs, RegistryEntry<Biome> biome) {
        sameBiomesToAdd[sameAs] << biome
    }

    void removeBiome(Biome biome) {
        removeBiome(biome.toRegistryEntry())
    }

    void removeBiome(RegistryEntry<Biome> biome) {
        biomesToRemove << biome
    }

    void removeBiome(RegistryKey<Biome> biomeKey) {
        biomeKeysToRemove << biomeKey
    }
}

interface IDimensionBiomesEnrichProvider {
    void enrichDimensionBiomes(DimensionBiomesEnrichContext context)
}